#if C_GRAPHICSMEDIA_LAB_GNU_GPL_C
#/*******************************************************************************
# * (C) Copyright 2003/2004 by Vladimir Vezhnevets <vvp@graphics.cs.msu.su> and *
# *                            Stanislav Soldatov  <stass@graphics.cs.msu.su>   *
# *******************************************************************************
# * $Id: PlugCVS.cpp,v 1.17 2004/04/04 23:57:04 stass Exp $
# *******************************************************************************
# * 
# *******************************************************************************
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
# *******************************************************************************/
#endif


#include "stdafx.h"
#include <io.h>
#include <process.h>
#include "PlugCVS.h"


///  
struct PluginStartupInfo g_Info;
struct FarStandardFunctions g_FSF;
char g_PluginRootKey[80];
BOOL g_bIsOldFar;

#define TIMEOUT              10

Operation_enum ShowOperationMenu(bool bCurDirOnly);
int AskConfirmation(Operation_enum in_Op, std::vector <std::string> &in_psFiles);
int AskTerminationConfirm(int in_iSeconds);
int CheckOut(Operation_enum mode);

//==========================================================================

void WINAPI _export SetStartupInfo(const struct PluginStartupInfo *in_pInfo)
{
  g_Info = *in_pInfo;
  g_bIsOldFar = TRUE;

  if(in_pInfo->StructSize >= sizeof(struct PluginStartupInfo))
  {
    g_FSF = *in_pInfo->FSF;
    g_Info.FSF = &::g_FSF;
    g_bIsOldFar = FALSE;
  }

}


//==========================================================================

HANDLE WINAPI OpenPlugin(int OpenFrom,int /*Item*/)
{
  std::vector <std::string>  psFiles;
  std::vector <const char *> ppcArgs;
  std::string pcFileName,
              pcModuleName;
  int iRes, iArgNum = 2;
  bool bCurDirOnly = false, bCheckTimeout = true, bNoConfirm = false;

  if(OpenFrom==OPEN_COMMANDLINE)
    return Browse();

  if(OpenFrom!=OPEN_PLUGINSMENU)
    return INVALID_HANDLE_VALUE;

  Operation_enum Operation;

  ///     
  if(!GetFileList(psFiles, bCurDirOnly))
    return INVALID_HANDLE_VALUE;

  ///  
  Operation = ShowOperationMenu(bCurDirOnly);

  ///   
  ppcArgs.resize(iArgNum);
  ppcArgs[0] = "cvs";

  switch (Operation)
  {
    case OP_LIGHTCOMMIT:
      ppcArgs[1] = "commit";
      PerformCommit(psFiles, false);
      return INVALID_HANDLE_VALUE;
      break;
    
    case OP_COMMIT:
      ppcArgs[1] = "commit";
      PerformCommit(psFiles, true);
      return INVALID_HANDLE_VALUE;
      break;

    case OP_ADD:
      ppcArgs[1] = "add";
      if (!GetAddParam(ppcArgs))
        return INVALID_HANDLE_VALUE;
      bNoConfirm = true;
      break;

    case OP_UPDATE:
      ppcArgs[1] = "update";
      if (!GetUpdateParam(ppcArgs))
        return INVALID_HANDLE_VALUE;
      bNoConfirm = true;
      break;

    case OP_REMOVE:
      ppcArgs[1] = "remove";
      break; 

    case OP_CHECKOUT:
    case OP_EXPORT:
      CheckOut(Operation);
      return INVALID_HANDLE_VALUE;

    case OP_SHOW_CHANGED:    
      ///          
      if (GetChangedList(".", psFiles, false))
        MarkFiles(psFiles);
      return INVALID_HANDLE_VALUE;
      break;

    case OP_SHOW_CONFLICTS:            
      ///          
      if (GetConflictsList(".", psFiles, false))
        MarkFiles(psFiles);
      return INVALID_HANDLE_VALUE;
      break;

    case OP_SHOW_UNKNOWN:            
      if (GetSpecialList(".", psFiles, std::string("?"), true, false))
        MarkFiles(psFiles);
      return INVALID_HANDLE_VALUE;
      break;

    case OP_SHOW_OLD:
      ///          
      if (GetSpecialList(".", psFiles, std::string("UP"), false, false))
        MarkFiles(psFiles);
      return INVALID_HANDLE_VALUE;
      break;

    case OP_LOG:            
      ppcArgs[1] = "log";
      break;

    case OP_STAT:            
      ppcArgs[1] = "stat"; 
      break;

    case OP_DIFF:            
      ppcArgs[1] = "diff";
      ppcArgs.push_back("-u");
      if (!GetDiffParam(ppcArgs))
        return INVALID_HANDLE_VALUE;
      bNoConfirm = true;
      break;

    case OP_BROWSE:            
      return Browse();

    default:
      return INVALID_HANDLE_VALUE;
      break;
  }



  iArgNum = ppcArgs.size();

  if (!bNoConfirm)
  {
    ///   
    iRes = AskConfirmation(Operation, psFiles);
    if (iRes)
      return INVALID_HANDLE_VALUE;
  }
   
  if(!bCurDirOnly)
  {
    ///    
    ppcArgs.resize(iArgNum + psFiles.size());

    if(!bCurDirOnly)
    {
      ///    
      ppcArgs.resize(iArgNum + psFiles.size());
  
      for (int i = 0; i < psFiles.size(); i++)  
        ppcArgs[i + iArgNum] = psFiles[i].c_str();
    }
    else
    {
      if(Operation == OP_ADD)
      ppcArgs.push_back("*");
    };

  }

  ppcArgs.push_back(NULL);


  pcFileName = CreateTempFileName();

  ///  CVS
  int iStatus = RunCVS(&ppcArgs[0], pcFileName.c_str(), bCheckTimeout);

  ///  
  ShowLogs(iStatus, pcFileName.c_str(), true);

  ///   
  _unlink(pcFileName.c_str());

  return (INVALID_HANDLE_VALUE);
}

enum CVSWorkingDialogItems
{
  CWI_CANCEL=-1,
  CWI_BREAK=1,
  CWI_TERMINATED=2
};
struct CVSThreadParam
{
  HANDLE  CVSProcess;
  HANDLE  Dialog;
  HANDLE  DieEvent;

};
void _cdecl CVSThread(void* arg)
{
  CVSThreadParam* tp=(CVSThreadParam*)arg;

  HANDLE Objects[2]={tp->CVSProcess,tp->DieEvent};

  long Result=WaitForMultipleObjects(sizeof(Objects)/sizeof(Objects[0]),Objects,
                                      FALSE, INFINITE); 

  if(Result==WAIT_OBJECT_0)
    g_Info.SendDlgMessage(tp->Dialog,DM_CLOSE,CWI_TERMINATED,0);

};

long WINAPI CVSDlgProc(
  HANDLE hDlg,
  int Msg,
  int Param1,
  long Param2
)
{
  static  CVSThreadParam* tp;
  switch(Msg)
  {
  case DN_INITDIALOG:
    {
      tp=(CVSThreadParam*)Param2;
      tp->Dialog=hDlg;
      tp->DieEvent=CreateEvent(NULL,FALSE,FALSE,NULL);
      if(_beginthread(CVSThread,0,(void*)tp))
        return TRUE;
      else
        return FALSE; //hmm
    }

  case DN_CLOSE:
      SetEvent(tp->DieEvent);
      //falldown to default dlgproc
  default:
    return g_Info.DefDlgProc(hDlg,Msg,Param1,Param2);
  }
}
//==========================================================================

int RunCVS(const char **in_ppcArgs, const char *in_pcLogFile, bool in_bCheckTimeout)
{
  int    iOldOut, iOldErr, iStatus, i, iTimeOut = TIMEOUT * 1000;
  DWORD  ExitCode;
  HANDLE Handle;
  std::vector <const char *> ppcArgs;

  if (!in_bCheckTimeout)
    iTimeOut = INFINITE;

  ppcArgs.push_back(in_ppcArgs[0]);

  i = 1;
  do
  {
    ppcArgs.push_back(in_ppcArgs[i]);
  }
  while (in_ppcArgs[i++]);

  ///  
  if (!in_pcLogFile || !in_ppcArgs)
    return -1000;

  ///      
  FILE *pFile = fopen(in_pcLogFile, "wt");

  if (!pFile)
    return -1001;

  ///   stderr  stdout
  iOldOut = _dup( 1 );   
  iOldErr = _dup( 2 );   

  ///  stderr  stdout   
  _dup2( _fileno(pFile), 1);
  _dup2( _fileno(pFile), 2);
  
  //int iTmp = fputs("\n", stdin);

  CVSThreadParam tp;
  static FarDialogItem CVSWorkingDialog []=
    {
      {DI_DOUBLEBOX,3,1,38,5,0,0,0,0,(char *)"CVS Running..."},
      {DI_BUTTON,0,4,0,0,0,0,DIF_CENTERGROUP,1,(char *)"Break"}
    };


  ///  cvs
  tp.CVSProcess = Handle = (void *)_spawnvp(_P_NOWAIT, "cvs", &ppcArgs[0]);

  int Result=g_Info.DialogEx(g_Info.ModuleNumber,
                  -1,-1,42,7,NULL,
                  CVSWorkingDialog,sizeof(CVSWorkingDialog)/sizeof(CVSWorkingDialog[0]),
                  0,0,
                  &CVSDlgProc,(long)&tp);

  if(Result==CWI_BREAK || Result==CWI_CANCEL)
      TerminateProcess( Handle, -168);




  ///  message
  //g_Info.Message(g_Info.ModuleNumber, FMSG_ALLINONE,
   //              NULL, (const char * const *)GetMsg(MSG_WORKING), 0, 0);



  ///  10 c...
  /*WaitStatus = WaitForSingleObject(Handle, iTimeOut); //TODO - plugin param

  g_Info.RestoreScreen(NULL);

  while (WaitStatus == WAIT_TIMEOUT)
  {
    int iRes = AskTerminationConfirm(10);
    if (!iRes)
      TerminateProcess( Handle, -168);

    g_Info.RestoreScreen(NULL);

    WaitStatus = WaitForSingleObject(Handle, iTimeOut); //TODO - plugin param
  }*/

  GetExitCodeProcess(Handle, &ExitCode);
  iStatus = (int)ExitCode;

  ///   (  )
  fclose(pFile);

  ///  stderr  stdout   
  _dup2( iOldOut, 1);
  _dup2( iOldErr, 2);

  return iStatus;
}

//==========================================================================
///     

bool GetFileList(std::vector <std::string> &out_psFiles, bool &io_bDirOnly)
{
  PanelInfo        PInfo;
  PluginPanelItem *pItem;

  ///    
  g_Info.Control(INVALID_HANDLE_VALUE, FCTL_GETPANELINFO, &PInfo);

  ///   ,   - 
  if (PInfo.PanelType != PTYPE_FILEPANEL)
    return false;

  /// ???   ,   -  FILEPANEL,   
  //if (PInfo.Plugin)
  //  return false;


  /// ,      
  pItem = &PInfo.SelectedItems[0];

  if (!PInfo.SelectedItemsNumber || io_bDirOnly
    || (PInfo.SelectedItemsNumber == 1 && strcmp(PInfo.SelectedItems[0].FindData.cFileName,"CVS") == 0) ||
       (PInfo.SelectedItemsNumber == 1 && !(PInfo.SelectedItems[0].Flags & PPIF_SELECTED)))
  {
    ///         -  
    out_psFiles.push_back(".");//PInfo.CurDir);
    io_bDirOnly = true;
  }
  else
  {
    int dir_len=strlen(PInfo.CurDir);
    ///     -  
    for (int i = 0; i < PInfo.SelectedItemsNumber; i++)      
    {
      const char* filename=PInfo.SelectedItems[i].FindData.cFileName;

      if(strnicmp(filename,PInfo.CurDir,dir_len)==0)
      {
        filename+=dir_len;
        if(filename[0]=='\\' || filename[0]=='/')
          filename++;
      };

      out_psFiles.push_back(filename);
    }
    io_bDirOnly = false;
  }

  return true;
}

//==========================================================================

bool GetChangedList(std::string in_sDir, std::vector<std::string> &out_psFiles,
                    bool in_bCurDirOnly)
{
  return GetSpecialList(in_sDir, out_psFiles, std::string("MA"), in_bCurDirOnly, false);
}

//==========================================================================

bool GetConflictsList(std::string in_sDir, std::vector<std::string> &out_psFiles,
                      bool in_bCurDirOnly)
{
  return GetSpecialList(in_sDir, out_psFiles, std::string("C"), in_bCurDirOnly, true);
}

//==========================================================================

bool GetSpecialList(std::string in_sDir, std::vector<std::string> &out_psFiles,
                    std::string in_sFlags, bool in_bCurDirOnly, bool in_bIgnoreReturn)
{
  std::string sTempFile, sBuf, sBuf2;
  int         iStatus;
  std::vector <const char *> ppcArgs;
  std::vector <std::string>  psArgs,
                             psRules;

  ///   
  for (int i = 0; i < in_sFlags.size(); i++)
  { 
    if (in_sFlags[i])
    {
      if (in_bCurDirOnly)
        psRules.push_back(" X %[^ \t\n/]");
      else
        psRules.push_back(" X%*[ ] %s");

      psRules[psRules.size() - 1][1] = in_sFlags[i];
    }
  }

  ///   
  ppcArgs.push_back("cvs");
  ppcArgs.push_back("-n");
  /// by stass:   
  ppcArgs.push_back("-q");
  ppcArgs.push_back("update");
  /// by stass:   -      
  if (in_bCurDirOnly)
    ppcArgs.push_back("-l");
  ppcArgs.push_back(in_sDir.c_str());
  ppcArgs.push_back(NULL);

  sTempFile = CreateTempFileName();

  ///  CVS
  iStatus = RunCVS(&ppcArgs[0], sTempFile.c_str(), true);

  ///   - 
  if (iStatus != 0 && !in_bIgnoreReturn)
  {
    ///  
    ShowLogs(iStatus, sTempFile.c_str(), true);

    ///   
    _unlink(sTempFile.c_str());
    return false;
  }

  /// Parse ,  
  out_psFiles.clear();

  FILE *pTempFile = fopen(sTempFile.c_str(), "rt");
  if (!pTempFile)
    return false;

  sBuf.resize(300);
  sBuf2.resize(300);

  while (!feof(pTempFile) && !ferror(pTempFile))
  {
    if (fgets(&sBuf[0], 300, pTempFile))
    {
      for (int i = 0; i < psRules.size(); i++)
      {
        if (sscanf(sBuf.c_str(), psRules[i].c_str(), &sBuf2[0]))
        {
          out_psFiles.push_back(sBuf2);
          break;
        }
      }
    }
  }

  fclose(pTempFile);
  ///   
  _unlink(sTempFile.c_str());

  return true;
}

//==========================================================================

bool MarkFiles(std::vector<std::string> &in_psFiles)
{
  PanelInfo        PInfo;
  PluginPanelItem *pItem;

  ///    
  g_Info.Control(INVALID_HANDLE_VALUE, FCTL_GETPANELINFO, &PInfo);

  ///   ,   - 
  if (PInfo.PanelType != PTYPE_FILEPANEL)
    return false;

  /// ???
  if (PInfo.Plugin)
    return false;

  ///      '/'
  for (int j = 0; j < in_psFiles.size(); j++)
  {
    int iPos = in_psFiles[j].find_first_of('/');
    if (iPos != std::string::npos)
      in_psFiles[j].resize(iPos);
  }

  ///     
  for (int i = 0; i < PInfo.ItemsNumber; i++)
  {
    /// ,      
    pItem = &PInfo.PanelItems[i];

    for (int j = 0; j < in_psFiles.size(); j++)
    {
      if (!strcmp(in_psFiles[j].c_str(), pItem->FindData.cFileName))
        break;
    }
    if (j < in_psFiles.size())
      pItem->Flags |= PPIF_SELECTED;
    else
      pItem->Flags &= ~PPIF_SELECTED;      
  }

  g_Info.Control(INVALID_HANDLE_VALUE, FCTL_SETSELECTION, &PInfo);
  g_Info.Control(INVALID_HANDLE_VALUE, FCTL_REDRAWPANEL, NULL);

  return true;
}


//==========================================================================

const char *GetMsg(int MsgId)
{
  return(g_Info.GetMsg(g_Info.ModuleNumber, MsgId));
}

//==========================================================================

bool PerformCommit(std::vector<std::string> &in_psFiles, bool in_bBuildList)
{
  std::string sTempFile, sCommentFile, sBuf, sBuf2;
  int         iStatus, iRes;
  std::vector <const char *> ppcArgs;
  std::vector <std::string>  psArgs,
                             psRules;
  std::vector <std::string>  psChangedFiles;
  bool                       bNonEmptyList = false;


  ///   
  sCommentFile = CreateTempFileName();

  FILE *pTempFile = fopen(sCommentFile.c_str(), "wt");
  if (!pTempFile)
    return false;


  ///     ( )
  if (in_bBuildList)
  {
    if (!GetChangedList(std::string("."), psChangedFiles, false))
      return false;

    ///       
    fprintf(pTempFile, ": ----------------------------------------------------------------------\n");
    fprintf(pTempFile, ": Modified and Added Files:\n");

    if (in_psFiles.size() == 1 && in_psFiles[0] == ".")
    {
      ///      ,  commit   
      for (int j = 0; j < psChangedFiles.size(); j++)
        fprintf(pTempFile, ":     %s\n", psChangedFiles[j].c_str());
      bNonEmptyList = psChangedFiles.size() > 0;
    }
    else
    {
      for (int i = 0; i < in_psFiles.size(); i++)
      {
        for (int j = 0; j < psChangedFiles.size(); j++)
        {
          if (strstr(psChangedFiles[j].c_str(), in_psFiles[i].c_str()))
          {
            fprintf(pTempFile, ":     %s\n", psChangedFiles[j].c_str());
            bNonEmptyList = true;
          }
        }
      }
    }
  }
  else
  {
    bNonEmptyList = true;
  }

  fprintf(pTempFile, ": ----------------------------------------------------------------------\n");
  fprintf(pTempFile, ": Dead Moroz CVS Plugin fo FAR Manager http://cgm.graphicon.ru/cvsplug  \n");
  fclose(pTempFile);

  if (bNonEmptyList)
  {
    ///   
    iStatus = g_Info.Editor(sCommentFile.c_str(), "CVS log", -1, -1, -1, -1, VF_DISABLEHISTORY, 0, 1);

    if (iStatus != EEC_MODIFIED)
      return false;

    ///   
    iRes = AskConfirmation(OP_COMMIT, in_psFiles);
    if (iRes)
      return false;

    ///   
    ppcArgs.push_back("cvs");
    ppcArgs.push_back("commit");
    ppcArgs.push_back("-F");
    ppcArgs.push_back(sCommentFile.c_str());
    for (int i = 0; i < in_psFiles.size(); i++)  
      ppcArgs.push_back(in_psFiles[i].c_str());
    ppcArgs.push_back(NULL);

    ///  CVS
    sTempFile = CreateTempFileName();
    iStatus = RunCVS(&ppcArgs[0], sTempFile.c_str(), true);

    ///   commit
    ShowLogs(iStatus, sTempFile.c_str(), true);

    ///   
    _unlink(sTempFile.c_str());
  }

  _unlink(sCommentFile.c_str());
  return true;
}

//===================================================================
//= Function name : CheckOut
//= Description   :    checkout  export   cvs
//= Return type   : int 
//===================================================================

int CheckOut( Operation_enum mode)
{

  char rev[32];
  rev[0]=0;rev[sizeof(rev)-1]=0;
  char date[32];
  date[0]=0;date[sizeof(date)-1]=0;
  char dest[128];
  dest[0]=0;dest[sizeof(dest)-1]=0;
  char module[128];
  module[0]=0;module[sizeof(module)-1]=0;
  char root[240];
  root[0]=0;root[sizeof(root)-1]=0;

  bool recursive;


  static FarDialogItem Dialog []=
    {
      {DI_DOUBLEBOX,3,1,72,12,0,0,0,0,0},
      {DI_BUTTON,0,11,0,0,0,0,DIF_CENTERGROUP,1,(char *)"OK"},
      {DI_BUTTON,0,11,0,0,0,0,DIF_CENTERGROUP,0,(char *)"Cancel"},
      {DI_TEXT,5,2,70,0,0,0,0,0,(char*)"&Module:"},
      {DI_EDIT,5,3,70,0,TRUE,(int)(const char*)"module", DIF_HISTORY|DIF_SELECTONENTRY,0,0},
      {DI_TEXT,5,4,70,0,0,0,0,0,(char*)"CVS &Root: (if empty CVS will use environment variable CVSROOT)"},
      {DI_EDIT,5,5,70,0,0,(int)(const char*)"cvsroot", DIF_HISTORY|DIF_USELASTHISTORY|DIF_SELECTONENTRY,0,0},
      {DI_TEXT,5,7,17,0,0,0,0,0,(char*)"Destinatio&n:"},
      {DI_EDIT,18,7,50,0,0,(int)(const char*)"destination", DIF_HISTORY|DIF_SELECTONENTRY,0,0},
      {DI_TEXT,5,9,17,0,0,0,0,0,(char*)"Revision/&Tag:"},
      {DI_EDIT,18,9,34,0,0,(int)(const char*)"revision", DIF_HISTORY|DIF_SELECTONENTRY,0,0},
      {DI_TEXT,37,9,41,0,0,0,0,0,(char*)"&Date:"},
      {DI_EDIT,42,9,50,0,0,(int)(const char*)"date", DIF_HISTORY,0,0},
      {DI_CHECKBOX,55,9,65,0,0,1,0,0,(char*)"R&ecursive"},
      {DI_TEXT,4,6,71,6,0,0,DIF_SEPARATOR,0,0},
      {DI_TEXT,4,10,71,10,0,0,DIF_SEPARATOR,0,0}
    };

    char* op=mode==OP_CHECKOUT?"Checkout":"Export";

    strcpy(Dialog[0].Data.Data,op);
    strcpy(Dialog[3].Data.Data,GetMsg(MSG_MODULENAME));
    //clean destination  - no default name!
    Dialog[8].Data.Data[0]=0;

    if(mode!=OP_CHECKOUT)
      strcpy(Dialog[12].Data.Data,"now");

    
    while(1)
    {
      int ExitCode = g_Info.Dialog(g_Info.ModuleNumber,
                            -1,-1,76,14,
                            NULL,Dialog,
                             sizeof(Dialog) / sizeof(Dialog[0]));

      if(ExitCode==-1 || ExitCode==2)
        return FALSE;
      if(ExitCode==1 && Dialog[4].Data.Data[0]==0)
      {
        g_Info.Message(g_Info.ModuleNumber,FMSG_WARNING|FMSG_MB_OK|FMSG_ALLINONE,
                       NULL, (const char * const *)"Error\nModule name is required", 0, 0);      
        continue;
      };
      break;
    };

    strncpy(module,Dialog[4].Data.Data,sizeof(module)-1);
    strncpy(root,Dialog[6].Data.Data,sizeof(root)-1);
    strncpy(dest,Dialog[8].Data.Data,sizeof(dest)-1);
    strncpy(rev,Dialog[10].Data.Data,sizeof(rev)-1);
    strncpy(date,Dialog[12].Data.Data,sizeof(date)-1);

    recursive=Dialog[13].Param.Selected!=0;


 
  int shift=1;
  const char** args=(const char**)new char*[10+1+1];
  args[0]="cvs";

  if(root[0])
  {
    args[shift]="-d";
    args[shift+1]=root;
    shift+=2;
  }
  args[shift++]=mode==OP_CHECKOUT?"checkout":"export";

  if(dest[0])
  {
    args[shift]="-d";
    args[shift+1]=dest;
    shift+=2;
  }

  if(rev[0])
  {
    args[shift]="-r";
    args[shift+1]=rev;
    shift+=2;
  }

  if(date[0])
  {
    args[shift]="-D";
    args[shift+1]=date;
    shift+=2;
  }

  if(!recursive)
    args[shift++]="-l";

  args[shift++]=module;
  args[shift++]=0;
        
  
  const char* fname = _tempnam(getenv("TEMP"), "cvsplug" );
  int iStatus = RunCVS(args, fname, false);
  //      
  ShowLogs(0 ,fname,true);
  _unlink(fname);

  
  delete[] args;

  return iStatus;
};


